#!/usr/bin/env python3

from trojanzoo.environ import env
from trojanzoo.models import Model
from trojanvision.datasets import ImageSet
from trojanvision.utils.model import Conv2d_SAME

import torch
import torch.nn as nn

from typing import Iterator
from collections.abc import Callable
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler


class _MagNet(nn.Module):
    """docstring for Model"""

    def __init__(self, structure: list[tuple[int, str]] = [3, 'average', 3],
                 activation: str = 'sigmoid', channel: int = 3, **kwargs):
        super().__init__()

        activation_fn = nn.ReLU()
        if activation == 'sigmoid':
            activation_fn = nn.Sigmoid()

        self.encoder = nn.Sequential()
        self.decoder = nn.Sequential()

        in_channels = channel
        for i, layer in enumerate(structure):
            match layer:
                case int():
                    conv = Conv2d_SAME(in_channels=in_channels,
                                       out_channels=structure[i],
                                       kernel_size=(3, 3))
                    in_channels = structure[i]
                    bn = nn.BatchNorm2d(structure[i])
                    self.encoder.add_module(f'conv{i+1:d}', conv)
                    self.encoder.add_module(f'bn{i+1:d}', bn)
                    self.encoder.add_module(f'{activation}{i+1:d}', activation_fn)
                case str():
                    module = nn.MaxPool2d(kernel_size=(2, 2)) if layer == 'max' \
                        else nn.AvgPool2d(kernel_size=(2, 2))
                    self.encoder.add_module('pool', module)
                case _:
                    raise TypeError(type(layer))

        for i, layer in enumerate(reversed(structure)):
            match layer:
                case int():
                    conv = Conv2d_SAME(in_channels=in_channels,
                                       out_channels=structure[i],
                                       kernel_size=(3, 3))
                    in_channels = structure[i]
                    bn = nn.BatchNorm2d(structure[i])
                    self.decoder.add_module(f'conv{i+1:d}', conv)
                    self.decoder.add_module(f'bn{i+1:d}', bn)
                    self.decoder.add_module(f'{activation}{i+1:d}', activation_fn)
                case str():
                    self.decoder.add_module('pool',
                                            nn.Upsample(scale_factor=(2, 2)))
                case _:
                    raise TypeError(type(layer))
        conv = Conv2d_SAME(structure[0], channel, kernel_size=(3, 3))
        bn = nn.BatchNorm2d(channel)
        self.decoder.add_module('conv', conv)
        self.decoder.add_module('bn', bn)
        self.decoder.add_module('activation', activation_fn)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x


class MagNet(Model):
    r"""MagNet proposed by Dongyu Meng from Shanghai Tech University in CCS 2017.
    It is an autoencoder for input images to defend against adversarial attacks.

    :Available model names:

        .. code-block:: python3

            ['magnet']

    See Also:
        * paper: `MagNet\: a Two-Pronged Defense against Adversarial Examples`_

    Args:
        structure (list[int | str]): The MagNet model structure.
            Defaults to

            * 1-channel images: ``[3, 'average', 3]`` (e.g, MNIST)
            * 3-channel images: ``[32]``
        activation (str): The activation layer in MagNet model.
            Choose from ``['sigmoid', 'relu']``.
            Defaults to ``'sigmoid'`` for 1-channel images (e.g, MNIST)
            and ``'relu'`` for 3-channel images.
        v_noise (float): The std of random Gaussian noise added to training data.
            Defaults to ``0.1``.

    .. _MagNet\: a Two-Pronged Defense against Adversarial Examples:
        https://arxiv.org/abs/1705.09064
    """
    available_models = ['magnet']

    def __init__(self, name: str = 'magnet',
                 dataset: ImageSet = None, model: type = _MagNet,
                 structure: list = None, activation: str = None,
                 v_noise: float = 0.1, **kwargs):
        self.v_noise: float = v_noise
        if structure is None:
            if dataset.data_shape[0] == 1:
                structure = [3, 'average', 3]
            else:
                structure = [32]
        if activation is None:
            if dataset.data_shape[0] == 1:
                activation = 'sigmoid'
            else:
                activation = 'relu'
        super().__init__(name=name, dataset=dataset, model=model,
                         structure=structure, activation=activation,
                         channel=dataset.data_shape[0], **kwargs)

    def get_data(self, data: tuple[torch.Tensor], v_noise: float = None,
                 mode='train') -> tuple[torch.Tensor, torch.Tensor]:
        _input = data[0]
        if mode == 'train':
            v_noise = v_noise if v_noise is not None else self.v_noise
            noise: torch.Tensor = v_noise * torch.rand_like(_input)
            data[0] = (_input + noise).clamp(0.0, 1.0)
            data[1] = _input.detach()
        else:
            data[0] = _input.detach()
            data[1] = _input.clone().detach()
        return data[0].to(device=env['device']), data[1].to(device=env['device'])

    def define_optimizer(
            self, parameters: str | Iterator[nn.Parameter] = 'full',
            OptimType: str | type[Optimizer] = 'Adam',
            lr: float = 0.1, momentum: float = 0.0, weight_decay: float = 1e-9,
            lr_scheduler: bool = True,
            lr_scheduler_type: str = 'CosineAnnealingLR',
            lr_step_size: int = 30, lr_gamma: float = 0.1,
            epochs: int = None, lr_min: float = 0.0,
            lr_warmup_epochs: int = 0, lr_warmup_method: str = 'constant',
            lr_warmup_decay: float = 0.01,
            **kwargs) -> tuple[Optimizer, _LRScheduler]:
        return super().define_optimizer(
            parameters=parameters, OptimType=OptimType,
            lr=lr, momentum=momentum, weight_decay=weight_decay,
            lr_scheduler=lr_scheduler, lr_scheduler_type=lr_scheduler_type,
            lr_step_size=lr_step_size, lr_gamma=lr_gamma,
            epochs=epochs, lr_min=lr_min,
            lr_warmup_epochs=lr_warmup_epochs,
            lr_warmup_method=lr_warmup_method,
            lr_warmup_decay=lr_warmup_decay,
            **kwargs)

    # define MSE loss function
    def define_criterion(
            self,
            **kwargs) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        entropy_fn = nn.MSELoss()

        def loss_fn(_output: torch.Tensor, _label: torch.Tensor):
            _output = _output.to(device=_label.device, dtype=_label.dtype)
            return entropy_fn(_output, _label)
        return loss_fn

    def accuracy(self, _output: torch.Tensor, _label: torch.Tensor,
                 num_classes: int = None, topk=(1, 5)):
        res = []
        for k in topk:
            res.append(-self.criterion(_output, _label))
        return res



#!/usr/bin/env python3

from .genotypes import Genotype
from .operations import get_op, PRIMITIVES

import torch
import torch.nn as nn
import numpy as np

from typing import Sequence
from collections.abc import Callable


class MixedOp(nn.Module):
    """ Mixed operation """

    def __init__(self, C: int, stride: int,
                 primitives: list[str] = PRIMITIVES,
                 **kwargs):
        super().__init__()
        self._ops = nn.ModuleList()
        for primitive in primitives:
            op = get_op(primitive, C, stride, affine=False, **kwargs)
            self._ops.append(op)

    def forward(self, x: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input
            weights: weight for each operation
        """
        return sum(w * op(x) for w, op in zip(weights, self._ops))


class Cell(nn.Module):
    def __init__(self, steps: int, multiplier: int,
                 C_prev_prev: int, C_prev: int, C: int,
                 reduction: bool, reduction_prev: bool,
                 **kwargs):
        super().__init__()
        self.reduction = reduction

        op_name = 'factorized_reduce' if reduction_prev else 'conv'
        self.preprocess0 = get_op(op_name, C_prev_prev, affine=False, C_out=C)
        self.preprocess1 = get_op('conv', C_prev, affine=False, C_out=C)

        self._steps = steps
        self._multiplier = multiplier

        self._ops = nn.ModuleList()
        for i in range(self._steps):
            for j in range(2 + i):
                stride = 2 if reduction and j < 2 else 1
                op = MixedOp(C, stride, **kwargs)
                self._ops.append(op)

    def forward(self, s0: torch.Tensor, s1: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
        s0 = self.preprocess0(s0)
        s1 = self.preprocess1(s1)
        states = [s0, s1]
        offset = 0
        for i in range(self._steps):
            s = sum(self._ops[offset + j](h, weights[offset + j]) for j, h in enumerate(states))
            offset += len(states)
            states.append(s)
        return torch.cat(states[-self._multiplier:], dim=1)


class FeatureExtractor(nn.Module):
    def __init__(self, C: int, layers: int,
                 steps: int = 4, multiplier: int = 4,
                 stem_multiplier: int = 3,
                 primitives: list[str] = PRIMITIVES,
                 **kwargs):
        super().__init__()
        self._C = C
        self._layers = layers
        self._steps = steps
        self._multiplier = multiplier
        self.primitives = primitives

        C_curr = stem_multiplier * C
        self.stem = nn.Sequential(
            nn.Conv2d(3, C_curr, 3, padding=1, bias=False),
            nn.BatchNorm2d(C_curr)
        )

        self.cells: Sequence[Cell] = nn.ModuleList()
        C_prev_prev, C_prev, C_curr = C_curr, C_curr, C
        reduction_prev = False
        for i in range(layers):
            reduction = False
            if i in [layers // 3, 2 * layers // 3]:
                C_curr *= 2
                reduction = True
            cell = Cell(steps, multiplier, C_prev_prev, C_prev, C_curr,
                        reduction, reduction_prev, primitives=primitives)
            reduction_prev = reduction
            self.cells.append(cell)
            C_prev_prev, C_prev = C_prev, multiplier * C_curr
        self.feats_dim = C_prev

        # k = sum(1 for i in range(self._steps) for n in range(2 + i))
        k = 2 * self._steps + ((self._steps - 1) * self._steps) // 2
        num_ops = len(self.primitives)
        if self._layers != 1:
            self.register_buffer('alphas_normal', 1e-3 * torch.randn(k, num_ops))
        self.register_buffer('alphas_reduce', 1e-3 * torch.randn(k, num_ops))
        self.alphas_normal: torch.Tensor  # = 1e-3 * torch.randn(k, num_ops)
        self.alphas_reduce: torch.Tensor  # = 1e-3 * torch.randn(k, num_ops)
        self.softmax: Callable[[torch.Tensor], torch.Tensor] = nn.Softmax(dim=-1)

    def arch_parameters(self) -> list[torch.Tensor]:
        if self._layers == 1:
            return [self.alphas_reduce]
        return [self.alphas_normal, self.alphas_reduce]

    def named_arch_parameters(self) -> list[tuple[str, torch.Tensor]]:
        if self._layers == 1:
            return [('alphas_reduce', self.alphas_reduce)]
        return [('alphas_normal', self.alphas_normal), ('alphas_reduce', self.alphas_reduce)]

    def forward(self, input):
        s0 = s1 = self.stem(input)
        for i, cell in enumerate(self.cells):
            weights = self.softmax(self.alphas_reduce if cell.reduction
                                   else self.alphas_normal)
            s0, s1 = s1, cell(s0, s1, weights)
        return s1

    def genotype(self) -> Genotype:
        gene_normal = self._parse(self.softmax(self.alphas_normal).detach().cpu().numpy())
        gene_reduce = self._parse(self.softmax(self.alphas_reduce).detach().cpu().numpy())

        concat = range(2 + self._steps - self._multiplier, self._steps + 2)
        genotype = Genotype(normal=gene_normal, normal_concat=concat,
                            reduce=gene_reduce, reduce_concat=concat)
        return genotype

    def _parse(self, weights: np.ndarray):
        gene: list[tuple[str, int]] = []
        n = 2
        start = 0
        for i in range(self._steps):
            end = start + n
            W = weights[start:end].copy()
            edges = sorted(range(i + 2), key=lambda x: -max(W[x][k]
                           for k in range(len(W[x]))
                           if k != self.primitives.index('none'))
                           )[:2]
            for j in edges:
                k_best = -1
                for k in range(len(W[j])):
                    if (k != self.primitives.index('none') and
                            (k_best == -1 or
                             W[j][k] > W[j][k_best])):
                        k_best = k
                gene.append((self.primitives[k_best], j))
            start = end
            n += 1
        return gene



#!/usr/bin/env python3

from trojanvision.models.imagemodel import _ImageModel, ImageModel
from trojanvision.utils.model_archs.pnasnet import PNASNetA, PNASNetB

import torch.nn as nn
from collections import OrderedDict


class _PNASNet(_ImageModel):

    def __init__(self, name: str = None, **kwargs):
        super().__init__(**kwargs)
        ModelClass = PNASNetA if '_a' in name else PNASNetB
        _model = ModelClass(num_classes=self.num_classes)
        self.features = nn.Sequential(OrderedDict([
            ('conv1', _model.conv1),
            ('bn1', _model.bn1),
            ('relu', nn.ReLU(inplace=True)),
            ('layer1', _model.layer1),
            ('layer2', _model.layer2),
            ('layer3', _model.layer3),
            ('layer4', _model.layer4),
            ('layer5', _model.layer5)
        ]))
        # self.pool = nn.AvgPool2d(8)
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', _model.linear)
        ]))


class PNASNet(ImageModel):
    r"""PNASNet proposed by Chenxi Liu from Johns Hopkins University in ECCV 2018.

    Note:
        The implementation is imported from a third-party github repo. The correctness can't be guaranteed.
        It might be better to reimplement according to tensorflow codes:
        https://github.com/tensorflow/models/blob/master/research/slim/nets/nasnet/pnasnet.py

    :Available model names:

        .. code-block:: python3

            ['pnasnet', 'pnasnet_a', 'pnasnet_b']

    See Also:
        * paper: `Progressive Neural Architecture Search`_
        * code: https://github.com/kuangliu/pytorch-cifar/blob/master/models/pnasnet.py

    .. _Progressive Neural Architecture Search:
        https://arxiv.org/abs/1712.00559
    """
    available_models = ['pnasnet', 'pnasnet_a', 'pnasnet_b']

    def __init__(self, name: str = 'pnasnet', layer: str = '_b',
                 model: type[_PNASNet] = _PNASNet, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)

    @classmethod
    def get_name(cls, name: str, layer: str = '') -> str:
        layer = layer if name == 'pnasnet' else ''
        return super().get_name(name, layer=layer)



#!/usr/bin/env python3

from trojanzoo.datasets import Dataset
from trojanvision.environ import env
from trojanvision.utils.transform import (get_transform_bit,
                                          get_transform_imagenet,
                                          get_transform_cifar,
                                          RandomMixup,
                                          RandomCutmix)

import torch
import torchvision.transforms as transforms
from torch.utils.data.dataloader import default_collate
import argparse
import os

from typing import TYPE_CHECKING
from typing import Iterable
from torchvision.datasets import VisionDataset  # TODO: python 3.10
import PIL.Image as Image
if TYPE_CHECKING:
    import torch.utils.data


class ImageSet(Dataset):
    r"""
    | The basic class representing an image dataset.
    | It inherits :class:`trojanzoo.datasets.Dataset`.

    Note:
        This is the implementation of dataset.
        For users, please use :func:`create` instead, which is more user-friendly.

    Args:
        norm_par (dict[str, list[float]]):
            Data normalization parameters of ``'mean'`` and ``'std'``
            (e.g., ``{'mean': [0.5, 0.4, 0.6], 'std': [0.2, 0.3, 0.1]}``).
            Defaults to ``None``.
        default_model (str): Default model for the dataset.
            Usually stored in the config.
            Defaults to ``'resnet18_comp'``.
        normalize (bool): Whether to use :any:`torchvision.transforms.Normalize`
            in dataset transform. Otherwise, use it as model preprocess layer.
        transform (str): The dataset transform type.

            * ``None |'none'`` (:any:`torchvision.transforms.PILToTensor`
              and :any:`torchvision.transforms.ConvertImageDtype`)
            * ``'bit'`` (transform used in BiT network)
            * ``'pytorch'`` (pytorch transform used in ImageNet training).

            Defaults to ``None``.

            Note:
                See :meth:`get_transform()` to get more details.
        auto_augment (bool): Whether to use
            :any:`torchvision.transforms.AutoAugment`.
            Defaults to ``False``.
        mixup (bool): Whether to use
            :class:`trojanvision.utils.transforms.RandomMixup`.
            Defaults to ``False``.
        mixup_alpha (float): :attr:`alpha` passed to
            :class:`trojanvision.utils.transforms.RandomMixup`.
            Defaults to ``0.0``.
        cutmix (bool): Whether to use
            :class:`trojanvision.utils.transforms.RandomCutmix`.
            Defaults to ``False``.
        cutmix_alpha (float): :attr:`alpha` passed to
            :class:`trojanvision.utils.transforms.RandomCutmix`.
            Defaults to ``0.0``.
        cutout (bool): Whether to use
            :class:`trojanvision.utils.transforms.Cutout`.
            Defaults to ``False``.
        cutout_length (int): Cutout length. Defaults to ``None``.
        **kwargs: keyword argument passed to
            :class:`trojanzoo.datasets.Dataset`.

    Attributes:
        data_type (str): Defaults to ``'image'``.
        num_classes (int): Defaults to ``1000``.
        data_shape (list[int]): The shape of image data ``[C, H, W]``.
            Defaults to ``[3, 224, 224]``.
    """

    name: str = 'imageset'
    data_type: str = 'image'
    num_classes = 1000
    data_shape = [3, 224, 224]

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
        r"""Add image dataset arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            The concrete dataset class may override this method to add more arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.

        See Also:
            :meth:`trojanzoo.datasets.Dataset.add_argument()`
        """
        super().add_argument(group)
        group.add_argument(
            '--dataset_normalize', dest='normalize', action='store_true',
            help='use transforms.Normalize in dataset transform. '
            '(It\'s used in model as the first layer by default.)')
        group.add_argument('--transform', choices=[None, 'none', 'bit', 'pytorch'])
        group.add_argument('--auto_augment', action='store_true',
                           help='use auto augment')
        group.add_argument('--mixup', action='store_true', help='use mixup')
        group.add_argument('--mixup_alpha', type=float, help='mixup alpha (default: 0.0)')
        group.add_argument('--cutmix', action='store_true', help='use cutmix')
        group.add_argument('--cutmix_alpha', type=float, help='cutmix alpha (default: 0.0)')
        group.add_argument('--cutout', action='store_true', help='use cutout')
        group.add_argument('--cutout_length', type=int, help='cutout length')
        return group

    def __init__(self, norm_par: dict[str, list[float]] = None,
                 default_model: str = 'resnet18_comp',
                 normalize: bool = False, transform: str = None,
                 auto_augment: bool = False,
                 mixup: bool = False, mixup_alpha: float = 0.0,
                 cutmix: bool = False, cutmix_alpha: float = 0.0,
                 cutout: bool = False, cutout_length: int = None,
                 **kwargs):
        self.norm_par: dict[str, list[float]] = norm_par
        self.normalize = normalize
        self.transform = transform
        self.auto_augment = auto_augment
        self.mixup = mixup
        self.mixup_alpha = mixup_alpha
        self.cutmix = cutmix
        self.cutmix_alpha = cutmix_alpha
        self.cutout = cutout
        self.cutout_length = cutout_length

        mixup_transforms = []
        if mixup:
            mixup_transforms.append(RandomMixup(self.num_classes, p=1.0, alpha=mixup_alpha))
        if cutmix:
            mixup_transforms.append(RandomCutmix(self.num_classes, p=1.0, alpha=cutmix_alpha))
        if len(mixup_transforms):
            mixupcutmix = mixup_transforms[0] if len(mixup_transforms) == 1 \
                else transforms.RandomChoice(mixup_transforms)

            def collate_fn(batch: Iterable[torch.Tensor]) -> Iterable[torch.Tensor]:
                return mixupcutmix(*default_collate(batch))  # noqa: E731
            self.collate_fn = collate_fn

        super().__init__(default_model=default_model, **kwargs)
        self.param_list['imageset'] = ['data_shape', 'norm_par',
                                       'normalize', 'transform',
                                       'auto_augment']
        if cutout:
            self.param_list['imageset'].append('cutout_length')

        if mixup:
            self.param_list['imageset'].append('mixup_alpha')
        if cutmix:
            self.param_list['imageset'].append('cutmix_alpha')

    def get_transform(self, mode: str, normalize: bool = None
                      ) -> transforms.Compose:
        r"""Get dataset transform based on :attr:`self.transform`.

            * ``None |'none'`` (:any:`torchvision.transforms.PILToTensor`
              and :any:`torchvision.transforms.ConvertImageDtype`)
            * ``'bit'`` (transform used in BiT network)
            * ``'pytorch'`` (pytorch transform used in ImageNet training).

        Args:
            mode (str): The dataset mode (e.g., ``'train' | 'valid'``).
            normalize (bool | None):
                Whether to use :any:`torchvision.transforms.Normalize`
                in dataset transform. Defaults to ``self.normalize``.

        Returns:
            torchvision.transforms.Compose: The transform sequence.
        """
        normalize = normalize if normalize is not None else self.normalize
        if self.transform == 'bit':
            return get_transform_bit(mode, self.data_shape)
        elif self.data_shape == [3, 224, 224]:
            transform = get_transform_imagenet(
                mode, use_tuple=self.transform != 'pytorch',
                auto_augment=self.auto_augment)
        elif self.transform != 'none' and self.data_shape in ([3, 16, 16], [3, 32, 32]):
            transform = get_transform_cifar(
                mode, auto_augment=self.auto_augment,
                cutout=self.cutout, cutout_length=self.cutout_length,
                data_shape=self.data_shape)
        else:
            transform = transforms.Compose([transforms.PILToTensor(),
                                            transforms.ConvertImageDtype(torch.float)])
        if normalize and self.norm_par is not None:
            transform.transforms.append(transforms.Normalize(
                mean=self.norm_par['mean'], std=self.norm_par['std']))
        return transform

    @staticmethod
    def get_data(data: tuple[torch.Tensor, torch.Tensor],
                 **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Process image data.
        Defaults to put input and label on ``env['device']`` with ``non_blocking``
        and transform label to ``torch.LongTensor``.

        Args:
            data (tuple[torch.Tensor, torch.Tensor]): Tuple of batched input and label.
            **kwargs: Any keyword argument (unused).

        Returns:
            (tuple[torch.Tensor, torch.Tensor]):
                Tuple of batched input and label on ``env['device']``.
                Label is transformed to ``torch.LongTensor``.
        """
        return (data[0].to(env['device'], non_blocking=True),
                data[1].to(env['device'], dtype=torch.long, non_blocking=True))

    def make_folder(self, img_type: str = '.png', **kwargs):
        r"""Save the dataset to ``self.folder_path``
        as :class:`trojanvision.datasets.ImageFolder` format.

        ``'{self.folder_path}/{self.name}/{mode}/{class_name}/{img_idx}.png'``

        Args:
            img_type (str): The image types to save. Defaults to ``'.png'``.
        """
        mode_list: list[str] = [
            'train', 'valid'] if self.valid_set else ['train']
        class_names = getattr(self, 'class_names',
                              [str(i) for i in range(self.num_classes)])
        for mode in mode_list:
            dataset: VisionDataset = self.get_org_dataset(mode, transform=None)
            class_counters = [0] * self.num_classes
            for image, target_class in list(dataset):
                image: Image.Image
                target_class: int
                class_name = class_names[target_class]
                _dir = os.path.join(
                    self.folder_path, self.name, mode, class_name)
                if not os.path.exists(_dir):
                    os.makedirs(_dir)
                image.save(os.path.join(
                    _dir, f'{class_counters[target_class]}{img_type}'))
                class_counters[target_class] += 1



#!/usr/bin/env python3

from ...abstract import BackdoorDefense


class RandomizedSmooth(BackdoorDefense):
    name: str = 'randomized_smooth'

    def detect(self, **kwargs):
        super().detect(**kwargs)
        self.model.randomized_smooth = True
        self.attack.validate_fn()



#!/usr/bin/env python3

from trojanzoo.defenses import Defense

from .advmind import AdvMind
from .curvature import Curvature
from .grad_train import GradTrain

__all__ = ['AdvMind', 'Curvature', 'GradTrain']

class_dict: dict[str, type[Defense]] = {
    'advmind': AdvMind,
    'curvature': Curvature,
    'grad_train': GradTrain,
}



#!/usr/bin/env python3

from ...abstract import BackdoorAttack

import torch
import argparse

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import torch.utils.data


class Unlearn(BackdoorAttack):

    name: str = 'unlearn'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--attack_source', help='attack source (default: "badnet")')
        group.add_argument('--mark_source', help='mark source (default: "attack")')
        return group

    def __init__(self, mark_source: str = 'attack', attack_source: str = 'badnet', **kwargs):
        self.attack_source = attack_source
        mark_source = mark_source if mark_source != 'attack' else attack_source
        self.mark_source = mark_source
        super().__init__(**kwargs)

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor],
                 keep_org: bool = True, poison_label=False, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        return super().get_data(data, keep_org=keep_org, poison_label=poison_label, **kwargs)

    def get_filename(self, **kwargs):
        return f'{self.attack_source}_{self.mark_source}_{self.train_mode}_' + super().get_filename(**kwargs)

    def get_poison_dataset(self, poison_label: bool = False, poison_num: int = None) -> torch.utils.data.Dataset:
        return super().get_poison_dataset(poison_label, poison_num)



#!/usr/bin/env python3

from .operations import PRIMITIVES, get_op

import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize
from torchvision import transforms

from typing import TYPE_CHECKING
from typing import Sequence
from .operations import Operation
if TYPE_CHECKING:
    pass


class Normalize(nn.Module):
    def __init__(self, temperature: float = 1.0):
        super().__init__()
        self.temperature = temperature

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        return X.div(self.temperature).softmax(0)


class MixedOp(nn.Module):
    """Mixed operation"""

    def __init__(self, primitives: list[str] = PRIMITIVES,
                 temperature: float = 0.05, **kwargs):
        super().__init__()
        self._ops: Sequence[Operation] = nn.ModuleList()
        for primitive in primitives:
            op = get_op(primitive, **kwargs)
            self._ops.append(op)
        self.temperature = temperature
        self.weights = nn.Parameter(torch.randn(len(primitives)))
        parametrize.register_parametrization(self, 'weights',
                                             Normalize(temperature))

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        result = torch.stack([op(_input) for op in self._ops])
        weights = self.weights.view([-1] + [1] * (result.dim() - 1))
        return (result * weights).sum(dim=0)

    def create_transform(self):
        return transforms.RandomChoice([op.create_transform()
                                        for op in self._ops],
                                       p=self.weights.tolist())


class SubPolicy(nn.Sequential):
    def __init__(self, operation_count: int = 4, **kwargs):
        super().__init__(*[MixedOp(**kwargs) for _ in range(operation_count)])

    def create_transform(self):
        return transforms.Compose([mixed_op.create_transform() for mixed_op in self])


class Policy(nn.Module):
    def __init__(self, num_sub_policies: int = 100, num_chunks: int = 8, **kwargs):
        super().__init__()
        self.num_sub_policies = num_sub_policies
        self.num_chunks = num_chunks
        self.sub_policies: Sequence[SubPolicy] = nn.ModuleList(
            [SubPolicy(**kwargs) for _ in range(num_sub_policies)])

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        input_list = _input.chunk(self.num_chunks if self.num_chunks != 0 else len(_input))
        idx_list = torch.randint(self.num_sub_policies, [len(input_list)])
        result = []
        for input_chunk, idx in zip(input_list, idx_list):
            result.append(self.sub_policies[idx](input_chunk))
        return torch.cat(result)

    def create_transform(self):
        return transforms.RandomChoice([sp.create_transform() for sp in self.sub_policies])



#!/usr/bin/env python3

from ...abstract import BackdoorDefense
from trojanvision.models import MagNet as MagNet_Model

import torch


class MagNet(BackdoorDefense):
    name: str = 'magnet'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.magnet = MagNet_Model(dataset=self.dataset, pretrained=True, folder_path=self.model.folder_path)

    def detect(self, **kwargs):
        super().detect(**kwargs)
        self.validate_fn()

    def validate_fn(self, **kwargs) -> tuple[float, float]:
        clean_acc, _ = self.model._validate(print_prefix='Validate Clean',
                                            get_data_fn=self.get_data, org=True, **kwargs)
        asr, _ = self.model._validate(print_prefix='Validate ASR',
                                      get_data_fn=self.get_data, keep_org=False, **kwargs)
        # self.model._validate(print_prefix='Validate Trigger Org',
        #                      get_data_fn=self.get_data, keep_org=False, poison_label=False, **kwargs)
        # print(f'Validate Confidence : {self.attack.validate_confidence():.3f}')
        return asr, clean_acc

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor],
                 **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.attack.get_data(data=data, **kwargs)
        _input = self.magnet(_input)
        return _input, _label



#!/usr/bin/env python3

import trojanzoo
from trojanzoo.utils.logger import MetricLogger

import torch
import torchvision
import torchvision.transforms.functional as F
from torchvision.transforms.functional import InterpolationMode

import math
import random
import skimage.metrics
import PIL.Image as Image

import argparse
import io
import os
import tarfile
from xml.etree.ElementTree import parse as ET_parse


sets: list[tuple[str, str]] = [('2012', 'train'), ('2012', 'val'), ('2007', 'train'), ('2007', 'val'), ('2007', 'test')]
NUM_ATTACK = 160
reflect_class = {'cat'}
background_class = {'person'}

norm = torch.distributions.normal.Normal(loc=0.0, scale=1.0)


def get_img_paths(datasets: list[torchvision.datasets.VOCDetection],
                  positive_class: set[str], negative_class: set[str]) -> list[str]:
    image_paths: list[str] = []
    for dataset in datasets:
        for index in range(len(dataset)):
            target = dataset.parse_voc_xml(ET_parse(dataset.annotations[index]).getroot())
            label_names: set[str] = {obj['name'] for obj in target['annotation']['object']}
            if len(positive_class & label_names) != 0 and len(negative_class & label_names) == 0:
                image_paths.append(dataset.images[index])
    return image_paths


def blend_images(background_img: torch.Tensor, reflect_img: torch.Tensor,
                 max_image_size: int = 560, ghost_rate: float = 0.49, alpha_t: float = None,
                 offset: tuple[int, int] = (0, 0), sigma: float = None, ghost_alpha: float = None):
    """
    Blend transmit layer and reflection layer together (include blurred & ghosted reflection layer) and
    return the blended image and precessed reflection image
    """
    if alpha_t is None:
        alpha_t = 1. - random.uniform(0.05, 0.45)
    h, w = background_img.shape[-2:]
    aspect_ratio = w / h
    h, w = (max_image_size, int(round(max_image_size * aspect_ratio))) if h > w \
        else (int(round(max_image_size / aspect_ratio)), max_image_size)
    # Original code uses cv2 INTER_CUBIC, which is slightly different from BICUBIC
    background_img = F.resize(background_img, size=(h, w), interpolation=InterpolationMode.BICUBIC).clamp(0, 1)
    reflect_img = F.resize(reflect_img, size=(h, w), interpolation=InterpolationMode.BICUBIC).clamp(0, 1)
    background_img.pow_(2.2)
    reflect_img.pow_(2.2)

    background_mask = alpha_t * background_img
    if random.random() < ghost_rate:
        # generate the blended image with ghost effect
        if ghost_alpha is None:
            ghost_alpha = abs(round(random.random()) - random.uniform(0.15, 0.5))
        if offset[0] == 0 and offset[1] == 0:
            offset = (random.randint(3, 8), random.randint(3, 8))
        reflect_1 = F.pad(background_img, [0, 0, offset[0], offset[1]])  # pad on right/bottom
        reflect_2 = F.pad(background_img, [offset[0], offset[1], 0, 0])  # pad on left/top
        reflect_ghost = ghost_alpha * reflect_1 + (1 - ghost_alpha) * reflect_2
        reflect_ghost = reflect_ghost[..., offset[0]: -offset[0], offset[1]: -offset[1]]
        reflect_ghost = F.resize(reflect_ghost, size=[h, w],
                                 interpolation=InterpolationMode.BICUBIC
                                 ).clamp(0, 1)  # no cubic mode in original code

        reflect_mask = (1 - alpha_t) * reflect_ghost
        reflection_layer = reflect_mask.pow(1 / 2.2)
    else:
        # generate the blended image with focal blur
        if sigma is None:
            sigma = random.uniform(1, 5)
        kernel_size = int(2 * math.ceil(2 * sigma) + 1)
        reflect_blur = F.gaussian_blur(reflect_img, kernel_size, sigma)
        blend = reflect_blur + background_img

        # get the reflection layers' proper range
        att = 1.08 + random.random() / 10.0
        mask = blend > 1
        mean = torch.tensor([blend[i, mask[i]].mean().nan_to_num(1.0).item()
                             for i in range(blend.size(0))]).view(-1, 1, 1)    # (C, 1, 1)
        reflect_blur = (reflect_blur - att * (mean - 1)).clamp(0, 1)

        def gen_kernel(kern_len: int = 100, nsig: int = 1) -> torch.Tensor:
            r"""Returns a 2D Gaussian kernel tensor."""
            interval = (2 * nsig + 1.) / kern_len
            x = torch.linspace(-nsig - interval / 2., nsig + interval / 2., kern_len + 1)
            # get normal distribution
            kern1d = norm.cdf(x).diff()
            kernel_raw = kern1d.outer(kern1d).sqrt()
            kernel = kernel_raw / kernel_raw.sum()  # TODO: is it auxiliary for positive numbers?
            kernel = kernel / kernel.max()
            return kernel
        h, w = reflect_blur.shape[-2:]
        new_h = random.randint(0, max_image_size - h - 10) if h < max_image_size - 10 else 0
        new_w = random.randint(0, max_image_size - w - 10) if w < max_image_size - 10 else 0
        g_mask = gen_kernel(max_image_size, 3).repeat(3, 1, 1)
        alpha_r = (1 - alpha_t / 2) * g_mask[..., new_h: new_h + h, new_w: new_w + w]

        reflect_mask = alpha_r * reflect_blur
        reflection_layer = (min(1., 4 * (1 - alpha_t)) * reflect_mask).pow(1 / 2.2)

    blended = (reflect_mask + background_mask).pow(1 / 2.2)
    background_layer = background_mask.pow(1 / 2.2)
    return blended, background_layer, reflection_layer


def read_tensor(fp: str) -> torch.Tensor:
    tensor = F.convert_image_dtype(F.pil_to_tensor(Image.open(fp)))
    return tensor.unsqueeze(0) if tensor.dim() == 2 else tensor


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--voc_root', default='~/voc')
    parser.add_argument('--tar_path', default='~/reflection.tar')
    kwargs = parser.parse_args().__dict__
    voc_root: str = kwargs['voc_root']
    tar_path: str = kwargs['tar_path']

    print('get image paths')
    datasets = [torchvision.datasets.VOCDetection(voc_root, year=year, image_set=image_set,
                                                  download=True) for year, image_set in sets]
    background_paths = get_img_paths(datasets, positive_class=background_class, negative_class=reflect_class)
    reflect_paths = get_img_paths(datasets, positive_class=reflect_class, negative_class=background_class)
    print()
    print('background: ', len(background_paths))
    print('reflect: ', len(reflect_paths))
    print()
    print('load images')
    reflect_imgs = [read_tensor(fp) for fp in reflect_paths]
    background_imgs = [read_tensor(fp) for i, fp in enumerate(background_paths) if i < NUM_ATTACK]

    print('writing tar file: ', tar_path)
    tf = tarfile.open(tar_path, mode='w')
    trojanzoo.environ.create(color=True, tqdm=True)
    logger = MetricLogger(meter_length=35)
    logger.create_meters(reflect_num='{count:3d}',
                         reflect_mean='{global_avg:.3f} ({min:.3f}  {max:.3f})',
                         diff_mean='{global_avg:.3f} ({min:.3f}  {max:.3f})',
                         blended_max='{global_avg:.3f} ({min:.3f}  {max:.3f})',
                         ssim='{global_avg:.3f} ({min:.3f}  {max:.3f})')
    candidates: set[int] = set()
    for background_img in logger.log_every(background_imgs):
        for i, reflect_img in enumerate(reflect_imgs):
            if i in candidates:
                continue
            blended, background_layer, reflection_layer = blend_images(background_img, reflect_img, ghost_rate=0.39)
            reflect_mean: float = reflection_layer.mean().item()
            diff_mean: float = (blended - reflection_layer).mean().item()
            blended_max: float = blended.max().item()
            logger.update(reflect_mean=reflect_mean, diff_mean=diff_mean, blended_max=blended_max)
            if reflect_mean < 0.8 * diff_mean and blended_max > 0.1:
                ssim: float = skimage.metrics.structural_similarity(
                    blended.numpy(), background_layer.numpy(), channel_axis=0)
                logger.update(ssim=ssim)
                if 0.7 < ssim < 0.85:
                    logger.update(reflect_num=1)
                    candidates.add(i)
                    filename = os.path.basename(reflect_paths[i])
                    bytes_io = io.BytesIO()
                    format = os.path.splitext(filename)[1][1:].lower().replace('jpg', 'jpeg')
                    F.to_pil_image(reflection_layer).save(bytes_io, format=format)
                    bytes_data = bytes_io.getvalue()
                    tarinfo = tarfile.TarInfo(name=filename)
                    tarinfo.size = len(bytes_data)
                    tf.addfile(tarinfo, io.BytesIO(bytes_data))
                    break
    tf.close()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

r"""
    TrojanVision config path:

    .. code-block:: python3

        config_path: dict[str, str] = {
            'package': os.path.dirname(__file__),   # trojanvision/configs/*/*.yml
            'user': os.path.normpath(os.path.expanduser('~/.trojanzoo/configs/trojanvision')),
            'project': os.path.normpath('./configs/trojanvision'),
        }

    Config class is defined in :class:`trojanzoo.configs.Config`.

    Warning:
        There is already a preset config instance ``trojanvision.configs.config``.

        NEVER call the class init method to create a new instance
        (unless you know what you're doing).
"""

import trojanzoo.configs
from trojanzoo.configs import Config

import os

config_path: dict[str, str] = {
    'package': os.path.dirname(__file__),   # trojanvision/configs/*/*.yml
    'user': os.path.normpath(os.path.expanduser('~/.trojanzoo/configs/trojanvision')),
    'project': os.path.normpath('./configs/trojanvision'),
}
config = Config(_base=trojanzoo.configs.config, **config_path)



#!/usr/bin/env python3

from ...abstract import ModelInspection
from trojanzoo.utils.tensor import repeat_to_batch
import torch
import torch.nn.functional as F

import argparse


class Tabor(ModelInspection):
    r"""Tabor proposed by Wenbo Guo and Dawn Song
    from Penn state and UC Berkeley in IEEE S&P 2019.

    It is a model inspection backdoor defense
    that inherits :class:`trojanvision.defenses.ModelInspection`.
    (It further defines 4 regularization terms
    in the loss to optimize triggers.)

    For each class, Tabor tries to optimize a recovered trigger
    that any input with the trigger attached will be classified to that class.
    If there is an outlier among all potential triggers, it means the model is poisoned.

    See Also:
        * paper: `TABOR\: A Highly Accurate Approach to Inspecting and Restoring Trojan Backdoors in AI Systems`_
        * code: https://github.com/UsmannK/TABOR

    Args:
        tabor_hyperparams (list[float]): List of weights for regularization terms.
            Defaults to ``[1e-6, 1e-5, 1e-7, 1e-8, 0, 1e-2]``

    .. _TABOR\: A Highly Accurate Approach to Inspecting and Restoring Trojan Backdoors in AI Systems:
        https://arxiv.org/abs/1908.01763
    """
    name: str = 'tabor'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--tabor_hyperparams', type=list,
                           help='list of weights for regularization terms '
                           '(default: [1e-6, 1e-5, 1e-7, 1e-8, 0, 1e-2])')
        return group

    def __init__(self, tabor_hyperparams: list[float] = [1e-6, 1e-5, 1e-7, 1e-8, 0, 1e-2], **kwargs):
        super().__init__(**kwargs)
        self.tabor_hyperparams = tabor_hyperparams

    def regularization_loss(self, _input: torch.Tensor, _label: torch.Tensor,
                            target: int):
        mark = self.attack.mark.mark
        # R1 - Overly large triggers
        mask_l1_norm = mark[-1].abs().sum()
        mask_l2_norm = mark[-1].square().sum()
        mask_r1 = mask_l1_norm + mask_l2_norm

        pattern_tensor: torch.Tensor = (1 - mark[-1]) * mark[:-1]
        pattern_l1_norm = pattern_tensor.abs().sum()
        pattern_l2_norm = pattern_tensor.square().sum()
        pattern_r1 = pattern_l1_norm + pattern_l2_norm

        # R2 - Scattered triggers
        pixel_dif_mask_col = F.mse_loss(mark[-1, :-1, :], mark[-1, 1:, :], reduction='sum')
        pixel_dif_mask_row = F.mse_loss(mark[-1, :, :-1], mark[-1, :, 1:], reduction='sum')
        mask_r2 = pixel_dif_mask_col + pixel_dif_mask_row

        pixel_dif_pat_col = F.mse_loss(pattern_tensor[:, :-1, :], pattern_tensor[:, 1:, :], reduction='sum')
        pixel_dif_pat_row = F.mse_loss(pattern_tensor[:, :, :-1], pattern_tensor[:, :, 1:], reduction='sum')
        pattern_r2 = pixel_dif_pat_col + pixel_dif_pat_row

        # R3 - Blocking triggers
        cropped_input_tensor: torch.Tensor = (1 - mark[-1]) * _input
        _cropped_output = self.model(cropped_input_tensor)
        r3 = self.model.criterion(_cropped_output, _label)

        # R4 - Overlaying triggers
        mask_crop_tensor = mark[-1] * mark[:-1]
        mask_crop_tensor = repeat_to_batch(mask_crop_tensor, _label.size(0))
        mask_cropped_output = self.model(mask_crop_tensor)
        r4 = self.model.criterion(mask_cropped_output, target * torch.ones_like(_label))

        loss = self.tabor_hyperparams[0] * mask_r1 + self.tabor_hyperparams[1] * pattern_r1 \
            + self.tabor_hyperparams[2] * mask_r2 + self.tabor_hyperparams[3] * pattern_r2 \
            + self.tabor_hyperparams[4] * r3 + self.tabor_hyperparams[5] * r4

        return loss

    def loss(self, _input: torch.Tensor, _label: torch.Tensor,
             target: int, trigger_output: torch.Tensor = None) -> torch.Tensor:
        return super().loss(_input, _label, target, trigger_output=trigger_output) \
            + self.regularization_loss(_input, _label, target)



#!/usr/bin/env python3

from trojanvision.models.imagemodel import ImageModel

from .natsbench import NATSbench

from .darts import DARTS
from .enas import ENAS
from .lanet import LaNet
from .pnasnet import PNASNet
from .proxylessnas import ProxylessNAS

__all__ = ['NATSbench', 'DARTS', 'ENAS', 'LaNet', 'PNASNet', 'ProxylessNAS']

class_dict: dict[str, type[ImageModel]] = {
    'nats_bench': NATSbench,
    'darts': DARTS,
    'enas': ENAS,
    'lanet': LaNet,
    'pnasnet': PNASNet,
    'proxylessnas': ProxylessNAS,
}
